#pragma once

#include <lektor/defines.h>
#include <lektor/mkv.h>
#include <lektor/uri.h>

#include <sqlite3.h>
#include <mpv/client.h>

#include <stdbool.h>
#include <linux/limits.h>

struct lkt_queue_state {
    int volume;
    int paused;
    int random;
    int repeat;
    int single;
    int current;
    int duration;
    int consume;
    int length;
};

/* Open correctly a database for lektor. */
bool database_new(sqlite3 **db);
bool database_open(sqlite3 *db, const char *dbpath);
bool database_attach(sqlite3 *db, const char *name, const char *dbpath);
bool database_detach(sqlite3 *db, const char *name);

/* Get information on the queue and currently playing kara. */
bool database_queue_state(sqlite3 *db, struct lkt_queue_state *res);
bool database_queue_current_kara(sqlite3 *db, struct kara_metadata *res, int *id);

bool database_queue_set_paused(sqlite3 *db, bool paused);
bool database_queue_set_current_index(sqlite3 *db, int idx);
bool database_queue_get_current_file(sqlite3 *db, char filepath[PATH_MAX]);

/* Mpv and database synchronisation. */
bool database_sync_mpv_state(sqlite3 *db, mpv_handle **mpv);

/* Update the database. */
bool database_update(sqlite3 *db, const char *kara_dir);
bool database_update_add(sqlite3 *db, const char *kara_path, struct kara_metadata *mdt, uint64_t id,
                         bool avail);

/* Control the content of the queue. */
bool database_queue_add_id(sqlite3 *db, int id, int priority);
bool database_queue_add_query(sqlite3 *db, const char *query, int priority);
bool database_queue_add_author(sqlite3 *db, const char *author, int priority);
bool database_queue_add_language(sqlite3 *db, const char *author, int priority);
bool database_queue_add_category(sqlite3 *db, const char *author, int priority);
bool database_queue_add_type(sqlite3 *db, const char *author, int priority);
bool database_queue_add_plt(sqlite3 *db, const char *plt_name, int priority);

bool database_queue_insert_query(sqlite3 *db, const char *query, int pos);
bool database_queue_insert_id(sqlite3 *db, int id);
bool database_queue_insert_author(sqlite3 *db, const char *author);
bool database_queue_insert_language(sqlite3 *db, const char *author);
bool database_queue_insert_category(sqlite3 *db, const char *author);
bool database_queue_insert_type(sqlite3 *db, const char *author);
bool database_queue_insert_plt(sqlite3 *db, const char *plt_name);

bool database_queue_del_id(sqlite3 *db, int id);
bool database_queue_del_pos(sqlite3 *db, int pos);
bool database_queue_clear(sqlite3 *db);
bool database_queue_crop(sqlite3 *db);
bool database_queue_move(sqlite3 *db, int from, int to);
bool database_queue_shuffle(sqlite3 *db);
bool database_queue_seekid(sqlite3 *db, int id, int *out_pos);

/* Control the playing state of the queue. */
bool database_queue_toggle_pause(sqlite3 *db);
bool database_queue_play(sqlite3 *db, int pos);
bool database_queue_stop(sqlite3 *db);

/* A search callback to be called after each matched row */
struct lkt_callback {
    bool (*call)(void *args, int id, int id_len, const char *sql_row);
    struct lkt_state *srv;
    size_t c;
    int iterations;
};

/* List the content of the queue */
bool database_queue_list(sqlite3 *db, size_t from, size_t to, struct lkt_callback *callback);

/* Search the database */
struct lkt_search {
    sqlite3 *db;
    sqlite3_stmt *stmt;
    enum lkt_search_type {
        lkt_search_database,
        lkt_search_playlist,
        lkt_search_queue,
    } type;
    void (*call)(void); /* Will be casted. */
    struct lkt_state *srv;
    size_t c;
    long continuation;  /* Is this a continuation from a previous command? */
    int msg_count;      /* How much messages we can send. */
};

typedef bool (*lkt_search_database_func)(struct lkt_state *srv, size_t c, int id, int id_len,
        const char *row);

bool database_search_database_init(sqlite3 *db, char *col_name, char *rgx, struct lkt_search *ret);
bool database_search_queue_init(sqlite3 *db, char *col_name, char *rgx, struct lkt_search *ret);
bool database_search_iter(struct lkt_search *item);

/* Next and prev operation on the queue. */
bool database_queue_next(sqlite3 *db, char filepath[PATH_MAX]);
bool database_queue_prev(sqlite3 *db, char filepath[PATH_MAX]);

/* Set a value in the config table */
bool database_config_set(sqlite3 *db, const char *section, const char *key, const char *value);
bool database_config_get_text(sqlite3 *db, const char *section, const char *key, char *value, size_t len);
bool database_config_get_int(sqlite3 *db, const char *section, const char *key, int *value);
bool database_config_exists(sqlite3 *db, const char *section, const char *key);
bool database_config_queue(sqlite3 *db, const char *option, int value);
bool database_config_queue_default(sqlite3 *db);

/* Get a value from the config table */
bool database_get_config(sqlite3 *db, const char *option, int *value);

/* Playlists */
struct lkt_playlist_metadata {
    const char *name;
    int last_update;
    int content_count;
};

/* Control playlists */
bool database_plt_create(sqlite3 *db, const char *name);
bool database_plt_remove(sqlite3 *db, const char *name);
bool database_plt_remove_pos(sqlite3 *db, const char *name, int pos);
bool database_plt_clear(sqlite3 *db, const char *name);
bool database_plt_rename(sqlite3 *db, const char *old_name, const char *new_name);

bool database_plt_export(sqlite3 *db, const char *name);
bool database_plt_import(sqlite3 *db, const char *name);

bool database_plt_add_uri(sqlite3 *db, const char *name, struct lkt_uri_t *uri);

/* User control, yeah, MPD authentification sucks. */
bool database_user_authentificate(sqlite3 *db, const char *password);
bool database_user_add(sqlite3 *db, const char *username, const char *password);

/* Stickers manipulations. */
struct sticker_callback {
    struct lkt_state *srv;
    size_t c;
    int uri, value, is_ok;
    const char *name;
    bool (*call)(void *args, const char *sticker, const char *type, int uri, int value);
};

bool database_sticker_create(sqlite3 *db, const char *name);
bool database_sticker_delete(sqlite3 *db, const char *name);
bool database_sticker_delete_specify(sqlite3 *sb, const char *type, int uri, const char *name);
bool database_sticker_list(sqlite3 *db, const char *type, struct sticker_callback *call);
bool database_sticker_set(sqlite3 *db, const char *type, const char *name, int uri, int value);
bool database_sticker_get(sqlite3 *db, const char *type, const char *name, int uri,
                          struct sticker_callback *call);
