#define _POSIX_C_SOURCE 200809L

#include <lektor/database.h>
#include <lektor/macro.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

/* Some notes:
   - There's one and only row.
   - `paused` 0 = play, 1 = paused
   - `random` whether the queue is played randomly
   - `repeat` whether the queue loops
   - `single` whether only one kara loops
   - `current` the position in the queue of the kara being played
   - `elapsed` the number of seconds from the beginning of the current kara
   - `duration` the total duration of the playing kara
   This schema is used to initialize the in-memory database. */
static const char *const SQL_MEM_SCHEM =
    "CREATE TABLE IF NOT EXISTS queue_state"
    "  ( id         INTEGER PRIMARY KEY DEFAULT 42 CHECK(id = 42)"
    "  , volume     INTEGER NOT NULL DEFAULT 100 CHECK(0 <= volume AND volume <= 100)"
    "  , paused     INTEGER NOT NULL DEFAULT 1"
    "  , random     INTEGER NOT NULL DEFAULT 0"
    "  , repeat     INTEGER NOT NULL DEFAULT 0"
    "  , single     INTEGER NOT NULL DEFAULT 0"
    "  , consume    INTEGER NOT NULL DEFAULT 0"
    "  , current    INTEGER CHECK(current > 0)"
    "  , duration   INTEGER CHECK(duration >= 0)"
    "  , elapsed    INTEGER CHECK(elapsed >= 0)"
    "  );\n"
    "INSERT INTO queue_state (id) VALUES (42);\n"
    "CREATE TABLE IF NOT EXISTS config"
    "  ( section    TEXT NOT NULL"
    "  , key        TEXT NOT NULL"
    "  , value      TEXT"
    "  , PRIMARY KEY (section, key)"
    "  ) WITHOUT ROWID;\n";

#define INVALID_CHARS_DBPATH    ":?!'\""
#define HEAP_LIMIT_SOFT         100 * 1024 * 1024
#define HEAP_LIMIT_HARD         150 * 1024 * 1024

int
is_sql_str_invalid(const char *str)
{
    size_t len = strlen(str);
    return strcspn(str, INVALID_CHARS_DBPATH) != len;
}

bool
database_new(sqlite3 **db)
{
    if (SQLITE_OK != sqlite3_enable_shared_cache(1)) {
        fprintf(stderr, " ! database_new: Failed to enable shared cache\n");
        return false;
    }

    if (sqlite3_soft_heap_limit64(HEAP_LIMIT_SOFT) < 0) {
        fprintf(stderr, " ! database_new: Failed to set soft heap limit to %u\n", HEAP_LIMIT_SOFT);
        return false;
    }

    if (sqlite3_hard_heap_limit64(HEAP_LIMIT_HARD) < 0) {
        fprintf(stderr, " ! database_new: Failed to set soft heap limit to %u\n", HEAP_LIMIT_HARD);
        return false;
    }

    /* SQLITE_OPEN_FULLMUTEX for serialized mode */
    static int flags = SQLITE_OPEN_READWRITE    |
                       SQLITE_OPEN_SHAREDCACHE  |
                       SQLITE_OPEN_NOFOLLOW     |
                       SQLITE_OPEN_NOMUTEX;

    if (SQLITE_OK != sqlite3_open_v2(":memory:", db, flags, NULL)) {
        fprintf(stderr, " ! database_open: Failed to open :memory: database\n");
        return false;
    }

    SQLITE_EXEC(*db, SQL_MEM_SCHEM, err_not_init);
    return true;
err_not_init:
    *db = NULL;
    return false;
}

static inline bool
__attach(sqlite3 *db, const char *name, const char *path)
{
    static const char SQL_ATTACH_TEMPLATE[] = "ATTACH '%s' AS '%s';";
    char SQL_ATTACH[LKT_MAX_SQLITE_STATEMENT];
    bool ret = false;

    snprintf(SQL_ATTACH, LKT_MAX_SQLITE_STATEMENT - 1, SQL_ATTACH_TEMPLATE, path, name);
    SQL_ATTACH[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_EXEC(db, SQL_ATTACH, err_no_attach);
    fprintf(stderr, " * Attached database '%s' with path '%s'\n", name, path);
    ret = true;
err_no_attach:
    return ret;
}

static inline bool
__detach(sqlite3 *db, const char *name)
{
    static const char SQL_DETACH_TEMPLATE[] = "DETACH '%s';\n";
    char SQL_DETACH[LKT_MAX_SQLITE_STATEMENT];
    bool ret = false;

    snprintf(SQL_DETACH, LKT_MAX_SQLITE_STATEMENT - 1, SQL_DETACH_TEMPLATE, name);
    SQL_DETACH[LKT_MAX_SQLITE_STATEMENT - 1] = 0;
    SQLITE_EXEC(db, SQL_DETACH, err_no_detach);
    fprintf(stderr, " * Detached database '%s'\n", name);
    ret = true;
err_no_detach:
    return ret;
}

static inline bool
__is_attached(sqlite3 *db, const char *name)
{
    static const char *SQL_STMT = "SELECT name FROM pragma_database_list WHERE name = ?;\n";
    sqlite3_stmt *stmt = 0;
    bool ret = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_ROW(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_open(sqlite3 *db, const char *dbpath)
{
    if (is_sql_str_invalid(dbpath)) {
        fprintf(stderr, " ! database_open: The database path %s is invalid\n", dbpath);
        return false;
    }

    return __attach(db, PROTECTED_DATABASE, dbpath);
}

bool
database_attach(sqlite3 *db, const char *name, const char *dbpath)
{
    if (!strcasecmp(PROTECTED_DATABASE, name)) {
        fprintf(stderr, " ! database_attach: The database " PROTECTED_DATABASE
                " is protected, can't attach a database with the same name\n");
        return false;
    }

    if (is_sql_str_invalid(name)) {
        fprintf(stderr, " ! database_attach: The database name %s is invalid\n", name);
        return false;
    }

    if (__is_attached(db, name)) {
        fprintf(stderr, " ! database_detach: Database named %s is already attached\n", name);
        return false;
    }

    if (!__attach(db, name, dbpath)) {
        fprintf(stderr, " ! database_attach: Failed to attach database named '%s' with path '%s'\n",
                name, dbpath);
        return false;
    }

    fprintf(stderr, " * Database %s attached\n", name);
    return true;
}

bool
database_detach(sqlite3 *db, const char *name)
{
    if (!strcasecmp(PROTECTED_DATABASE, name)) {
        fprintf(stderr, " ! database_detach: The database " PROTECTED_DATABASE
                " is protected, can't detach it\n");
        return false;
    }

    if (is_sql_str_invalid(name)) {
        fprintf(stderr, " ! database_detach: The database name %s is invalid\n", name);
        return false;
    }

    if (!__is_attached(db, name)) {
        fprintf(stderr, " ! database_detach: Database named %s is not attached\n", name);
        return false;
    }

    if (!__detach(db, name)) {
        fprintf(stderr, " ! database_detach: Failed to detach database named %s\n", name);
        return false;
    }

    fprintf(stderr, " * Database %s detached\n", name);
    return true;
}
