#define _POSIX_C_SOURCE 200809L

#include <lektor/macro.h>
#include <lektor/defines.h>
#include <lektor/net.h>
#include <lektor/cmd.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <signal.h>
#include <locale.h>
#include <stdnoreturn.h>
#include <assert.h>
#include <stdbool.h>
#include <stdarg.h>
#include <limits.h>

#define LKT_KEY_VALUE_SEP           ": \n\t\0"
#define STR_MATCH(str1, str2)       (! strcasecmp(str1, str2))
#define STR_NMATCH(str1, str2, n)   (! strncasecmp(str1, str2, n))

/* The help. */

noreturn void
help(void)
{
    static const char *help_str =
        "USAGE: lkt [OPTIONS] <COMMAND> [ARGS [...]]\n"
        "\n"
        "  OPTIONS:\n"
        "    host         named of the lektor's host, can be resolved.\n"
        "    port         port on which lektor is listening.\n"
        "\n"
        "    options most be passed as one word (no spaced), such as the following:\n"
        "    % lkt host=sakura port=6601 play\n"
        "\n"
        "  COMMANDS:\n"
        "    help           display this help message.\n"
        "    play [?idx]    toggle play/pause state of lektor, may start at a certain index.\n"
        "    stop           stop the playback but not the window if it exists.\n"
        "    status         get the status of lektor.\n"
        "    current        get the currently playing song.\n"
        "    prev           play previous kara in the queue.\n"
        "    next           play the next kara in the queue.\n"
        "    shuffle        shuffle lektor's playlist and play it from the begening.\n"
        "    queue          tge queue sub command.\n"
        "    plt            the playlist sub command.\n"
        "    search         the search sub command.\n"
        "\n"
        "  QUEUE COMMANDS:\n"
        "    pos <arg>      the qrgument can be a position or a range.\n"
        "    <count>        prints the next songs in the queue.\n"
        "    pop [pos]      pop a song in the queue.\n"
        "    add <query>    add a kara to the queue with a query.\n"
        "    seek <query>   seek the next song that matchs the query.\n"
        "    delete <id>    delete the song by its id in the queue.\n"
        "    clear          clear the queue.\n"
        "    crop           crop the queue.\n"
        "\n"
        "  PLAYLIST COMMANDS:\n"
        "    create <plt>               creates a playlist after the argument.\n"
        "    destroy <plt>              delete a playlist after the argument.\n"
        "    add <plt> <type> <query>   add something to a playlist from the database.\n"
        "    delete <plt> <id>          delete domething from a playlist.\n"
        "\n"
        "  SEARCH COMMANDS:\n"
        "    get <query>            prints the results of the query.\n"
        "    add <query>            add to the queue and prints the results of the query.\n"
        "    insert <query>         insert on top of the aueue and prints the results of the query.\n"
        "    plt <plt> <query>      search inside a playlist and prints the results.\n"
        "    count <query>          count the number of songs that matches the query.\n"
        "    queue <query>          prints kara that matches from the queue.\n"
        "\n"
        "  QUERY:\n"
        "    A query is passed in argument of a COMMAND and is composed of:\n"
        "     - The first word must be the type\n"
        "     - the rest is used for the sqlite regex\n"
        "    Supported types are: title, [a]ny, source, [auth]or, [lang]uage, type, title\n"
        "\n"
        "  RANGE:\n"
        "    A range is specified like BEGIN:END which implies from BEGIN to END included\n"
        "\n";
    write(1, help_str, strlen(help_str));
    exit(EXIT_SUCCESS);
}

noreturn void
help__(struct lkt_cmd_args *args)
{
    (void) args;
    help();
}

static noreturn inline void
fail(const char *message)
{
    fprintf(stderr, "%s\n", message);
    exit(EXIT_FAILURE);
}

/* Type definition. */

typedef struct {
    const char *host;   /* Serveur host, may be resolved.   */
    const char *port;   /* Serveur port.                    */
    const char **argv;  /* Pointer to the argv from the main
                           function.                        */
    int argc;           /* Argument count of the args.      */
} args_t;

/* Global variables. */

static const char *host;
static const char *port;
static const char *LKT_QUEUE_DEFAULT[] = { "10" };

/* Communication functions and fonction that interact with lektor stuff. */

static int
lkt_valid_type(const char *type)
{
    return (STR_MATCH(type, "all")      ||
            STR_MATCH(type, "any")      ||
            STR_MATCH(type, "a")        ||
            STR_MATCH(type, "id")       ||
            STR_MATCH(type, "title")    ||
            STR_MATCH(type, "type")     ||
            STR_MATCH(type, "cat")      ||
            STR_MATCH(type, "category") ||
            STR_MATCH(type, "author")   ||
            STR_MATCH(type, "auth")     ||
            STR_MATCH(type, "source")   ||
            STR_MATCH(type, "lang")     ||
            STR_MATCH(type, "language"));
}

static int
read_socket(FILE *sock, char *buff, size_t max_len)
{
    size_t i, len;

    for (i = 0; i < max_len; ++i) {
        len = fread(buff + i, sizeof(char), 1, sock);

        if (buff[i] == '\n' || len != 1)
            goto end;
    }

end:
    return i;
}

static inline int
write_socket(FILE *sock, const char *buff, size_t len)
{
    int ret = 1;

    if (fwrite(buff, sizeof(char), len, sock) != len)
        goto end;

    ret = 0;
end:
    return ret;
}

static int
write_socket_format(FILE *sock, const char *format, ...)
{
    int ret = 1, size;
    char buff[LKT_MESSAGE_MAX];
    va_list ap;

    va_start(ap, format);
    size = vsnprintf(buff, LKT_MESSAGE_MAX - 1, format, ap);
    buff[LKT_MESSAGE_MAX - 1] = 0;

    if (size < 0)
        goto err;

    if (fwrite(buff, sizeof(char), size, sock) < (unsigned int) size)
        goto err;

    ret = 0;
err:
    va_end(ap);
    return ret;
}

static FILE *
create_socket(const char *host, const char *port)
{
    FILE *socket_desc;
    struct addrinfo hints, *found;
    socklen_t sock_len;
    struct sockaddr sock;

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_flags = 0;
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;

    assert(! getaddrinfo(host, port, &hints, &found));

    sock = *found->ai_addr;
    sock_len = found->ai_addrlen;
    freeaddrinfo(found);

    int cx = socket(AF_INET, SOCK_STREAM, 0);

    if (cx <= 0 || connect(cx, &sock, sock_len))
        exit(EXIT_FAILURE);

    assert(socket_desc = fdopen(cx, "r+"));

    return socket_desc;
}

static FILE *
lkt_connect(void)
{
    char buff[LKT_MESSAGE_MAX];
    size_t recv_len;

    FILE *sock = create_socket(host, port);
    recv_len = read_socket(sock, buff, LKT_MESSAGE_MAX);

    assert(recv_len > 0);

    return sock;
}

static noreturn inline void
lkt_send_and_exit(const char *buffer, size_t len)
{
    exit(write_socket(lkt_connect(), buffer, len));
}

static char *
lkt_skip_key(char *buffer)
{
    size_t len = strcspn(buffer, LKT_KEY_VALUE_SEP);
    char *ret = &buffer[len + 1];
    len = strspn(ret, ": ");
    return ret + len;
}

/* Functions implementing options. */

noreturn void
queue_clear__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the clear command takes no arguments");
    static const char cmd__[] = "clear\nclose\n";
    lkt_send_and_exit(cmd__, sizeof(cmd__));
}

noreturn void
queue_crop__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the crop command takes no arguments");
    static const char cmd__[] = "crop\nclose\n";
    lkt_send_and_exit(cmd__, sizeof(cmd__));
}


noreturn void
next__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the next command takes no arguments");
    static const char cmd__[] = "next\nclose\n";
    lkt_send_and_exit(cmd__, sizeof(cmd__));    /* In bytes. */
}

noreturn void
prev__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the previous command takes no arguments");
    static const char cmd__[] = "previous\nclose\n";
    lkt_send_and_exit(cmd__, sizeof(cmd__));    /* In bytes. */
}

noreturn void
stop__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the stop command takes no arguments");
    static const char cmd__[] = "stop\nclose\n";
    lkt_send_and_exit(cmd__, sizeof(cmd__));    /* In bytes. */
}

noreturn void
play__(struct lkt_cmd_args *args)
{
    if (args->argc > 1)
        fail("Invalid argument, the play command takes no arguments or only one");

    int pos = 0;
    if (args->argc != 0)
        pos = atoi(args->argv[0]);

    static const char status__[]        = "status\n";
    static const char cmd_play__[]      = "play\nclose\n";
    static const char cmd_play_from__[] = "play %d\nclose\n";
    static const char cmd_pause__[]     = "pause\nclose\n";

    char buff[LKT_MESSAGE_MAX];

    FILE *sock = lkt_connect();

    if (write_socket(sock, status__, sizeof(status__))) /* In bytes. */
        goto error;

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        if (read_socket(sock, buff, LKT_MESSAGE_MAX - 1) <= 0)
            exit(EXIT_FAILURE);

        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);

        if (STR_NMATCH(buff, "state", len)) {
            fclose(sock);

            if (STR_NMATCH(lkt_skip_key(buff), "stop", 4)) {
                if (!pos)
                    lkt_send_and_exit(cmd_play__, sizeof(cmd_play__));  /* In bytes. */
                else
                    exit(write_socket_format(lkt_connect(), cmd_play_from__, pos));
            }

            else
                lkt_send_and_exit(cmd_pause__, sizeof(cmd_pause__));    /* In bytes. */

            goto error;
        }
    }

error:
    exit(EXIT_FAILURE);
}

noreturn void
current__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the current command takes no arguments");

    static const char current_song__[] = "currentsong\n";
    char buff[LKT_MESSAGE_MAX];

    char *mem = NULL;
    FILE *sock = lkt_connect();

    if (write_socket(sock, current_song__, sizeof(current_song__))) /* In bytes. */
        goto error;

    assert(mem = calloc(6 * LKT_MESSAGE_MAX, sizeof(char)));
    assert(memset(mem, 0, 6 * LKT_MESSAGE_MAX * sizeof(char)));

    char *const title    = &mem[0];
    char *const author   = &mem[LKT_MESSAGE_MAX];
    char *const source   = &mem[2 * LKT_MESSAGE_MAX];
    char *const type     = &mem[3 * LKT_MESSAGE_MAX];
    char *const category = &mem[4 * LKT_MESSAGE_MAX];
    char *const language = &mem[5 * LKT_MESSAGE_MAX];

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        if (read_socket(sock, buff, LKT_MESSAGE_MAX - 1) <= 0)
            goto error;

        const size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);
        char *value = lkt_skip_key(buff);
        const size_t value_size = strcspn(value, "\n\0") * sizeof(char);

        if (! strncasecmp("title", buff, len))
            assert(memcpy(title, value, value_size));

        if (! strncasecmp("author", buff, len))
            assert(memcpy(author, value, value_size));

        if (! strncasecmp("source", buff, len))
            assert(memcpy(source, value, value_size));

        if (!strncasecmp("type", buff, len))
            assert(memcpy(type, value, value_size));

        if (!strncasecmp("language", buff, len))
            assert(memcpy(language, value, value_size));

        if (!strncasecmp("category", buff, len))
            assert(memcpy(category, value, value_size));

        /* At this point every key has been parsed. */
        if (! strncmp(buff, "OK", 2))
            goto ok;
        else if (! strncmp(buff, "ACK", 3))
            goto error;
    }

ok:
    if (language[0])
        printf("%s - %s / %s - %s - %s [ %s ]\n",
               category, language, source, type, title, author);
    else
        printf("%s / %s - %s - %s [ %s ]\n",
               category, source, type, title, author);

    exit(EXIT_SUCCESS);
error:
    exit(EXIT_FAILURE);
}

noreturn void
queue_pop__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the status command takes no arguments");

    int ret = EXIT_FAILURE, songid = 0;
    char buff[LKT_MESSAGE_MAX];
    FILE *sock = lkt_connect();

    /* Get lektor's status. */

    if (write_socket(sock, "status\n", sizeof("status\n")))
        goto error;

#define assign_int(str, var) if (! strncmp(buff, str, len)) { var = (atoi(lkt_skip_key(buff))); continue; }

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        if (read_socket(sock, buff, LKT_MESSAGE_MAX - 1) <= 0)
            goto error;

        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);
        assign_int("songid", songid)

        /* At this point every key has been parsed. */
        if (! strncmp(buff, "OK", 2))
            break;
        else if (! strncmp(buff, "ACK", 3))
            goto error;
    }

    fclose(sock);
    sock = lkt_connect();
    if (!songid)
        goto error;
    write_socket_format(sock, "next\ndeleteid %d\n", songid);
    ret = EXIT_SUCCESS;
error:
    exit(ret);
}

noreturn void
status__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the status command takes no arguments");

    static const char *const status_str__ = "status\n";
    int ret = EXIT_FAILURE, it = 0;

    char buff[LKT_MESSAGE_MAX];
    char flags[24];

    bool play = false, stopped = true;
    int time_pos = 0, time_duration = 0, song_index = 0, plt_len = 0;

    memset(flags, 0, 24 * sizeof(char));
    FILE *sock = lkt_connect();

    /* Get lektor's status. */

    if (write_socket(sock, status_str__, strlen(status_str__)))
        goto error;

#define assign_flag(str, f)  if (! strncmp(buff, str, len) && (atoi(lkt_skip_key(buff)) == 1)) { flags[it++] = f; continue; }
#define assign_int(str, var) if (! strncmp(buff, str, len)) { var = (atoi(lkt_skip_key(buff))); continue; }

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        if (read_socket(sock, buff, LKT_MESSAGE_MAX - 1) <= 0)
            goto error;

        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);

        if (! strncmp(buff, "state", len)) {
            char *it = lkt_skip_key(buff);
            play = ! strncmp(it, "play", 4);
            stopped = ! strncmp(it, "stop", 4);
            continue;
        }

        assign_flag("random", 'r')
        assign_flag("repeat", 'l')
        assign_flag("single", 's')
        assign_flag("consume", 'c')

        assign_int("elapsed", time_pos)
        assign_int("duration", time_duration)
        assign_int("song", song_index)
        assign_int("playlistlength", plt_len)

        /* At this point every key has been parsed. */
        if (! strncmp(buff, "OK", 2))
            goto print;
        else if (! strncmp(buff, "ACK", 3))
            goto error;
    }

print:
    fclose(sock);

    printf("Lektor: %s\nQueue: #%d/%d\nPlayback: %d/%ds\nFlags: %s\n",
           play ? "play" : "stopped",
           song_index + 1, plt_len, time_pos, time_duration,
           flags[0] ? flags : "(none)");

    /* If there is a kara loaded in mpv, get this kara. */
    if (!stopped) {
        printf("Kara: ");
        current__(args);
    }

    ret = EXIT_SUCCESS;
error:
    exit(ret);
}

noreturn void
shuffle__(struct lkt_cmd_args *args)
{
    if (args->argc != 0)
        fail("Invalid argument, the shuffle command takes no arguments");

    static const char *const shuffle_str__ = "shuffle\nplay\nclose\n";
    lkt_send_and_exit(shuffle_str__, strlen(shuffle_str__));
}

noreturn void
queue_delete__(struct lkt_cmd_args *args)
{
    if (args->argc != 1)
        fail("Invalid argument, need onlt one argument");

    static const char *cmd_id__ = "deleteid %d\nclose\n";
    int dumy = 0;
    FILE *sock = lkt_connect();
    char buff[3];

    sscanf(args->argv[0], "%d", &dumy);
    if (dumy != 0) {
        write_socket_format(sock, cmd_id__, dumy);
        goto check;
    }

    fail("Invalid argument");

check:
    assert(read_socket(sock, buff, 2) > 0);
    if (buff[0] == 'O' && buff[1] == 'K')
        exit(EXIT_SUCCESS);
    else
        exit(EXIT_FAILURE);
}

noreturn void
queue_add__(struct lkt_cmd_args *args)
{
    char buff[LKT_MESSAGE_MAX];
    int i;

    if (args->argc < 1)
        fail("Invalid argument, the add command should takes at least two arguments");

    if (!lkt_valid_type(args->argv[0]))
        fail("Invalid argument, the type given to the add command is invalid");

    FILE *sock = lkt_connect();

    write_socket(sock, "add ", 4 * sizeof(char));
    write_socket(sock, args->argv[0], strlen(args->argv[0]));
    write_socket(sock, "://", 3 * sizeof(char));

    for (i = 1; i < args->argc - 1; ++i) {
        write_socket(sock, args->argv[i], strlen(args->argv[i]));
        write_socket(sock, " ", sizeof(char));
    }

    /* Here we have `i == argc - 1`. */
    write_socket(sock, args->argv[i], strlen(args->argv[i]));
    write_socket(sock, "\n", sizeof(char));

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        assert(read_socket(sock, buff, LKT_MESSAGE_MAX - 1) > 0);

        if (! strncmp(buff, "OK", 2))
            exit(EXIT_SUCCESS);

        else if (! strncmp(buff, "ACK", 3))
            exit(EXIT_FAILURE);
    }
}

noreturn void
queue_seek__(struct lkt_cmd_args *args)
{
    if (args->argc != 1)
        fail("The seek command needs one argument");

    char *endptr, buf[3];
    long id = strtol(args->argv[0], &endptr, 0);

    if ((errno == ERANGE && (id == LONG_MAX || id == LONG_MIN)) ||
        (errno != 0 && id == 0)                                 ||
        (endptr == args->argv[0]))
        fail("Invalid argument, not an integer");

    if (*endptr != '\0')
        fail("Invalid argument, must be only one integer");

    FILE *sock = lkt_connect();
    write_socket_format(sock, "playid %ld\n", id);
    read_socket(sock, buf, 2);
    exit(!strncmp(buf, "OK", 2));
}

noreturn void
queue_pos__(struct lkt_cmd_args *args)
{
    char buff[LKT_MESSAGE_MAX], *endptr;

    if (args->argc != 1)
        fail("Invalid argument for the pos command");

    long continuation = 0, up = 0;

    continuation = strtol(args->argv[0], &endptr, 0);
    if ((errno == ERANGE && (continuation == LONG_MAX || continuation == LONG_MIN)) || (errno != 0
            && continuation == 0) || (endptr == args->argv[0]))
        fail("Invalid argument, not an integer");

    if (*endptr != '\0') {
        /* A range */
        if (*(++endptr) == '\0')
            fail("Invalid argument, a range is two integers");
        up = atoi(endptr);
    }


    FILE *sock = NULL;
redo:
    sock = lkt_connect();
    if (up != 0)
        write_socket_format(sock, "playlist %d:%d\n", continuation, up);
    else
        write_socket_format(sock, "playlist %d\n", continuation);

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        assert(read_socket(sock, buff, LKT_MESSAGE_MAX - 1) > 0);

        if (! strncmp(buff, "continue:", strlen("continue:"))) {
            continuation = atoi(lkt_skip_key(buff));
            if (continuation > 0) {
                fclose(sock);
                goto redo;
            }
        }

        if (! strncmp(buff, "OK", 2))
            exit(EXIT_SUCCESS);
        else if (! strncmp(buff, "ACK", 3))
            exit(EXIT_FAILURE);

        fprintf(stdout, "%s", buff);
    }
}

noreturn void
queue_list__(struct lkt_cmd_args *args)
{
    char buff[LKT_MESSAGE_MAX], *endptr;
    FILE *sock = NULL;
    long continuation = 0, song_index = 1;

    /* Arguments stuff. */

    if (args->argc == 0)
        args->argv = LKT_QUEUE_DEFAULT;
    else if (args->argc != 1)
        fail("Invalid argument");

    continuation = strtol(args->argv[0], &endptr, 0);
    if ((errno == ERANGE && (continuation == LONG_MAX || continuation == LONG_MIN)) || (errno != 0
            && continuation == 0) || (endptr == args->argv[0]))
        fail("Invalid argument, not an integer");

    if (*endptr != '\0')
        fail("Invalid argument");

    /* Get the current pos to get limits for the playlist command. */

    sock = lkt_connect();
    if (write_socket(sock, "status\n", sizeof("status\n")))
        fail("Communication error");

#define assign_int(str, var) if (! strncmp(buff, str, len)) { var = (atoi(lkt_skip_key(buff))); continue; }

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        if (read_socket(sock, buff, LKT_MESSAGE_MAX - 1) <= 0)
            fail("Connextion error");

        size_t len = strcspn(buff, LKT_KEY_VALUE_SEP);
        assign_int("song", song_index);

        /* At this point every key has been parsed. */
        if (! strncmp(buff, "OK", 2))
            break;
        else if (! strncmp(buff, "ACK", 3))
            exit(EXIT_FAILURE);
    }
    fclose(sock);

    /* Get the content of the queue. */

    continuation = labs(continuation);
    song_index = MAX(song_index + 1, 1);
    snprintf(buff, LKT_MESSAGE_MAX - 1, "%ld:%ld", song_index, song_index + continuation - 1);
    buff[LKT_MESSAGE_MAX - 1] = '\0';
    args->argc = 1;
    args->argv[0] = buff;
    queue_pos__(args);
}

/* Functions implementing options, but for for playlists. */

noreturn void
plt_add__(struct lkt_cmd_args *args)
{
    int i;
    char buff[2];
    FILE *sock = lkt_connect();

    if (args->argc < 3)
        fail("Invalid argument, need at least three arguments");

    if (!lkt_valid_type(args->argv[1]))
        fail("Invalid argument, type for the query is invalid");

    write_socket(sock, "playlistadd ", strlen("playlistadd "));
    write_socket(sock, args->argv[0], strlen(args->argv[0]));
    write_socket(sock, " ", sizeof(char));
    write_socket(sock, args->argv[1], strlen(args->argv[1]));
    write_socket(sock, "://", 3 * sizeof(char));

    for (i = 2; i < args->argc - 1; ++i) {
        write_socket(sock, args->argv[i], strlen(args->argv[i]));
        write_socket(sock, " ", sizeof(char));
    }

    /* Here we have `i == argc - 1`. */
    write_socket(sock, args->argv[i], strlen(args->argv[i]));
    write_socket(sock, "\n", sizeof(char));

    assert(read_socket(sock, buff, 2) > 0);
    if (buff[0] == 'O' && buff[1] == 'K')
        exit(EXIT_SUCCESS);
    else
        exit(EXIT_FAILURE);
}

noreturn void
plt_delete__(struct lkt_cmd_args *args)
{
    FILE *sock = lkt_connect();
    char buff[2];
    int i;

    if (args->argc < 1)
        fail("Invalid argument, need at least two arguments");

    write_socket(sock, "playlistdelete ", strlen("playlistdelete "));
    write_socket(sock, args->argv[0], strlen(args->argv[0]));
    write_socket(sock, " ", sizeof(char));

    for (i = 1; i < args->argc - 1; ++i) {
        write_socket(sock, args->argv[i], strlen(args->argv[i]));
        write_socket(sock, " ", sizeof(char));
    }

    /* Here we have `i == argc - 1`. */
    write_socket(sock, args->argv[i], strlen(args->argv[i]));
    write_socket(sock, "\n", sizeof(char));

    assert(read_socket(sock, buff, 2) > 0);
    if (buff[0] == 'O' && buff[1] == 'K')
        exit(EXIT_SUCCESS);
    else
        exit(EXIT_FAILURE);
}

noreturn void
plt_destroy__(struct lkt_cmd_args *args)
{
    if (args->argc != 1)
        fail("Invalid argument, you may only specify the name of the playlist to delete");

    FILE *sock = lkt_connect();
    char buff[2];
    const char *name = args->argv[0];

    write_socket(sock, "rm ", 3 * sizeof(char));
    write_socket(sock, name, strlen(name));
    write_socket(sock, "\n", sizeof(char));

    assert(read_socket(sock, buff, 2) > 0);
    if (buff[0] == 'O' && buff[1] == 'K')
        exit(EXIT_SUCCESS);
    else
        exit(EXIT_FAILURE);
}

noreturn void
plt_create__(struct lkt_cmd_args *args)
{
    if (args->argc != 1)
        fail("Invalid argument, you may only specify the name of the playlist to create");

    FILE *sock = lkt_connect();
    char buff[2];
    const char *name = args->argv[0];

    write_socket(sock, "playlistadd ", strlen("playlistadd "));
    write_socket(sock, name, strlen(name));
    write_socket(sock, "\n", sizeof(char));

    assert(read_socket(sock, buff, 2) > 0);
    if (buff[0] == 'O' && buff[1] == 'K')
        exit(EXIT_SUCCESS);
    else
        exit(EXIT_FAILURE);
}

/* Search functions. */

noreturn void
search_with_cmd__(struct lkt_cmd_args *args, const char *cmd)
{
    if (args->argc < 2)
        fail("Invalid number of arguments, need at least a valid query");

    if (!lkt_valid_type(args->argv[0]))
        fail("Invalid type for the query");

    char buff[LKT_MESSAGE_MAX];
    int continuation = 0, i;
    FILE *sock = NULL;
redo:
    sock = lkt_connect();

    write_socket_format(sock, "%d %s", continuation, cmd);
    for (i = 0; i < args->argc; ++i)
        write_socket_format(sock, " %s", args->argv[i]);
    write_socket(sock, "\n", sizeof("\n") / sizeof(char));

    for (;;) {
        memset(buff, 0, LKT_MESSAGE_MAX * sizeof(char));
        assert(read_socket(sock, buff, LKT_MESSAGE_MAX - 1) > 0);

        if (! strncmp(buff, "continue:", strlen("continue:"))) {
            continuation = atoi(lkt_skip_key(buff));
            fclose(sock);
            goto redo;
        }

        if (! strncmp(buff, "OK", 2))
            exit(EXIT_SUCCESS);
        else if (! strncmp(buff, "ACK", 3))
            exit(EXIT_FAILURE);

        fprintf(stdout, "%s", buff);
    }
}

noreturn void
search_get__(struct lkt_cmd_args *args)
{
    search_with_cmd__(args, "search");
}

noreturn void
search_add__(struct lkt_cmd_args *args)
{
    search_with_cmd__(args, "searchadd");
}

noreturn void
search_insert__(struct lkt_cmd_args *args)
{
    (void) args;
    fail("Not implemented");
}

noreturn void
search_plt__(struct lkt_cmd_args *args)
{
    search_with_cmd__(args, "listplaylist");
}

noreturn void
search_count__(struct lkt_cmd_args *args)
{
    search_with_cmd__(args, "count");
}

noreturn void
search_queue__(struct lkt_cmd_args *args)
{
    search_with_cmd__(args, "playlistfind");
}

/* Parsing stuff. */

static struct lkt_cmd_opt options_queue[] = {
    { .name = "pos",        .call = queue_pos__     },
    { .name = "pop",        .call = queue_pop__     },
    { .name = "add",        .call = queue_add__     },
    { .name = "seek",       .call = queue_seek__    },
    { .name = "delete",     .call = queue_delete__  },
    { .name = "clear",      .call = queue_clear__   },
    { .name = "crop",       .call = queue_crop__    },
    LKT_OPT_DEFAULT(queue_list__),
};

static struct lkt_cmd_opt options_plt[] = {
    { .name = "help",       .call = help__          },
    { .name = "add",        .call = plt_add__       },
    { .name = "delete",     .call = plt_delete__    },
    { .name = "destroy",    .call = plt_destroy__   },
    { .name = "create",     .call = plt_create__    },
    LKT_OPT_NULL,
};

static struct lkt_cmd_opt options_search[] = {
    { .name = "help",       .call = help__          },
    { .name = "get",        .call = search_get__    },
    { .name = "add",        .call = search_add__    },
    { .name = "insert",     .call = search_insert__ },
    { .name = "plt",        .call = search_plt__    },
    { .name = "count",      .call = search_count__  },
    { .name = "queue",      .call = search_queue__  },
    LKT_OPT_NULL,
};

noreturn void
queue__(struct lkt_cmd_args *args)
{
    if (args->argc == 0)
        queue_list__(args);

    lkt_cmd_parse(options_queue, args->argc, args->argv, help);
}

noreturn void
search__(struct lkt_cmd_args *args)
{
    if (args->argc == 0)
        fail("Invalid argument, you must specify a sub command for the search command");

    lkt_cmd_parse(options_search, args->argc, args->argv, help);
}

noreturn void
plt__(struct lkt_cmd_args *args)
{
    if (args->argc == 0)
        fail("Invalid argument, you must specify a sub command for the playlist command");

    lkt_cmd_parse(options_plt, args->argc, args->argv, help);
}

static struct lkt_cmd_opt options_[] = {
    { .name = "help",       .call = help__    },
    { .name = "current",    .call = current__ },
    { .name = "play",       .call = play__    },
    { .name = "next",       .call = next__    },
    { .name = "previous",   .call = prev__    },
    { .name = "queue",      .call = queue__   },
    { .name = "shuffle",    .call = shuffle__ },
    { .name = "status",     .call = status__  },
    { .name = "stop",       .call = stop__    },
    { .name = "plt",        .call = plt__     },
    { .name = "search",     .call = search__  },
    LKT_OPT_NULL,
};

/* The sigpipe function, if SIGPIPE signal is sent. */

static void
sigpipe__(int sig)
{
    (void) sig;
    static const char *errmsg = "Exit because sigpipe\n";
    write(2, errmsg, strlen(errmsg));
    exit(EXIT_FAILURE);
}

/* Functions declarations. */

static void
parse_args(args_t *args, int argc, const char **argv)
{
    int i, got = 0;
    size_t len;

    for (i = 1; i < argc && i < 3; ++i) {
        len = strcspn(argv[i], "=");

        if (STR_NMATCH("host", argv[i], len)) {
            args->host = (argv[i] + len + 1);
            ++got;
        }

        else if (STR_NMATCH("port", argv[i], len)) {
            args->port = (argv[i] + len + 1);
            ++got;
        }
    }

    args->argv = &argv[got + 1];
    args->argc = argc - (got + 1);
}

/* Functions that make a command happen on lektor. */

/* Playlist functions. */

int
main(int argc, const char **argv)
{
    assert(NULL != setlocale(LC_ALL, "en_US.UTF-8"));   /* BECAUSE!         */
    assert(!signal(SIGPIPE, sigpipe__));                /* Argument checks. */

    args_t args = {
        .host = "localhost",
        .port = "6600",
    };

    parse_args(&args, argc, argv);

    host = args.host;
    port = args.port;

    /* Communication with lektor. */
    lkt_cmd_parse(options_, args.argc, args.argv, help);
}
