#define _POSIX_C_SOURCE 200809L

#include <lektor/thread.h>
#include <pthread.h>
#include <sys/types.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

struct __args {
    void *(*start)(struct lkt_thread_arg *);
    struct lkt_thread_arg *arg;
};

static void *
__start(void *args__)
{
    struct __args *args = (struct __args *) args__;
    void *(*start)(struct lkt_thread_arg *) = args->start;
    struct lkt_thread_arg *arg = args->arg;
    free(args__);
    return start(arg);
}

int
lkt_th_new(struct lkt_thread *th, unsigned int init_sizes,
           void *(*func)(struct lkt_thread_arg *), void *args)
{
    int ret = 1;
    pthread_mutex_t mtx1 = PTHREAD_MUTEX_INITIALIZER;
    pthread_mutex_t mtx2 = PTHREAD_MUTEX_INITIALIZER;
    struct lkt_thread th_ = {
        .initial_size   = init_sizes,
        .input_lock     = mtx1,
        .output_lock    = mtx2,
        .input_len      = 0,
        .output_len     = 0,
        .input_size     = init_sizes,
        .output_size    = init_sizes,
        .input_cells    = calloc(init_sizes, sizeof(void *)),
        .output_cells   = calloc(init_sizes, sizeof(void *)),
    };

    if (!th_.input_cells || !th_.output_cells)
        goto out_of_memory;

    *th = th_;
    struct __args *__args = calloc(1, sizeof(struct __args));
    __args->start = func;
    __args->arg = args;
    __args->arg->self = th;

    ret = pthread_create(&(th->th), NULL, __start, __args);

    if (ret)
        goto end;

end:
    fprintf(stderr, " %c lkt_th_new: %s\n",
            ret ? '!' : '*',
            ret ? "Failed to create a new thread" : "Created a new thread");
    return ret;

out_of_memory:
    fprintf(stderr, " * lkt_th_new: Out of memory\n");
    if (th_.input_cells)
        free((void *) th_.input_cells);
    if (th_.output_cells)
        free((void *) th_.output_cells);
    ret = errno = ENOMEM;
    goto end;
}

int
lkt_th_join(struct lkt_thread *th, void **ret)
{
    int sta = pthread_join(th->th, ret);

    if (sta)
        fprintf(stderr, " ! lkt_th_join: Failed to join thread\n");

    if (th->input_cells)
        free((void *) th->input_cells);
    if (th->output_cells)
        free((void *) th->output_cells);

    if (!sta)
        fprintf(stderr, " * lkt_th_join: Thread joined\n");

    memset(th, 0, sizeof(struct lkt_thread));
    return sta;
}

static inline int
th_append(unsigned int *len, unsigned int *size, void ***cells,
          pthread_mutex_t *lock, void *ptr)
{
    void *new;
    int ret = 0;

    if (pthread_mutex_lock(lock)) {
        fprintf(stderr, " ! th_append: Failed to lock\n");
        goto end;
    }

    if (*len == *size) {
        new = realloc((void *) *cells, (*len + *size) * sizeof(void *));

        if (new == NULL) {
            fprintf(stderr, " ! th_append: Out of memory\n");
            ret = (errno = ENOMEM);
            goto end;
        }

        (*len) += *size;
        (*cells) = new;
    }

    (*cells)[(*len)++] = ptr;
    ret = 0;
end:
    if (pthread_mutex_unlock(lock)) {
        fprintf(stderr, " ! th_append: Failed to lock\n");
        ret = 1;
    }

    return ret;
}

static inline int
th_pop(unsigned int *len, void **cells, pthread_mutex_t *lock, void **ptr)
{
    int ret = 1;

    if (pthread_mutex_lock(lock)) {
        fprintf(stderr, " ! th_pop: Failed to lock\n");
        goto end;
    }

    if (*len > 0)
        *ptr = cells[--(*len)];
    else
        *ptr = NULL;

    ret = 0;
end:

    if (pthread_mutex_unlock(lock)) {
        fprintf(stderr, " ! th_pop: Failed to lock\n");
        ret = 1;
    }

    return ret;
}

static inline int
th_head(unsigned int len, void **cells, pthread_mutex_t *lock, void **ptr)
{
    int ret = 1;

    if (pthread_mutex_lock(lock)) {
        fprintf(stderr, " ! th_pop: Failed to lock\n");
        goto end;
    }

    if (len > 0)
        *ptr = cells[len - 1];
    else
        *ptr = NULL;

    ret = 0;
end:

    if (pthread_mutex_unlock(lock)) {
        fprintf(stderr, " ! th_pop: Failed to lock\n");
        ret = 1;
    }

    return ret;
}

int
lkt_th_append_input(struct lkt_thread *th, void *ptr)
{
    return th_append((unsigned int *) &th->input_len, (unsigned int *) &th->input_size,
                     (void ***) &th->input_cells, &th->input_lock, ptr);
}

int
lkt_th_append_output(struct lkt_thread *th, void *ptr)
{
    return th_append((unsigned int *) &th->output_len, (unsigned int *) &th->output_size,
                     (void ***) &th->output_cells, &th->output_lock, ptr);
}

int
lkt_th_pop_input(struct lkt_thread *th, void **ptr)
{
    return th_pop((unsigned int *) &th->input_len, (void **) th->input_cells,
                  &th->input_lock, ptr);
}


int
lkt_th_pop_output(struct lkt_thread *th, void **ptr)
{
    return th_pop((unsigned int *) &th->output_len, (void **) th->output_cells,
                  &th->output_lock, ptr);
}

int
lkt_th_head_input(struct lkt_thread *th, void **ptr)
{
    return th_head((unsigned int) th->input_len, (void **) th->input_cells, &th->input_lock, ptr);
}

int
lkt_th_head_output(struct lkt_thread *th, void **ptr)
{
    return th_head((unsigned int) th->output_len, (void **) th->output_cells, &th->output_lock, ptr);
}
